"""
Generate a undirected graph
"""
from argparse import ArgumentParser, FileType
import graphviz
import json
from graphviz_overlay.overlays.er import EntityRelationshipOverlay
from graphviz_overlay.overlays.graphviz import GraphOverlay, DigraphOverlay
import sys


valid_attrs = [
    '_background',
    'area',
    'arrowhead',
    'arrowsize',
    'arrowtail',
    'bb',
    'bgcolor',
    'center',
    'charset',
    'class',
    'clusterrank',
    'color',
    'colorscheme',
    'comment',
    'compound',
    'concentrate',
    'constraint',
    'Damping',
    'decorate',
    'defaultdist',
    'dim',
    'dimen',
    'dir',
    'diredgeconstraints',
    'distortion',
    'dpi',
    'edgehref',
    'edgetarget',
    'edgetooltip',
    'edgeURL',
    'epsilon',
    'esep',
    'fillcolor',
    'fixedsize',
    'fontcolor',
    'fontname',
    'fontnames',
    'fontpath',
    'fontsize',
    'forcelabels',
    'gradientangle',
    'group',
    'head_lp',
    'headclip',
    'headhref',
    'headlabel',
    'headport',
    'headtarget',
    'headtooltip',
    'headURL',
    'height',
    'href',
    'id',
    'image',
    'imagepath',
    'imagepos',
    'imagescale',
    'inputscale',
    'K',
    'label',
    'label_scheme',
    'labelangle',
    'labeldistance',
    'labelfloat',
    'labelfontcolor',
    'labelfontname',
    'labelfontsize',
    'labelhref',
    'labeljust',
    'labelloc',
    'labeltarget',
    'labeltooltip',
    'labelURL',
    'landscape',
    'layer',
    'layerlistsep',
    'layers',
    'layerselect',
    'layersep',
    'layout',
    'len',
    'levels',
    'levelsgap',
    'lhead',
    'lheight',
    'lp',
    'ltail',
    'lwidth',
    'margin',
    'maxiter',
    'mclimit',
    'mindist',
    'minlen',
    'mode',
    'model',
    'mosek',
    'newrank',
    'nodesep',
    'nojustify',
    'normalize',
    'notranslate',
    'nslimit',
    'nslimit1',
    'ordering',
    'orientation',
    'outputorder',
    'overlap',
    'overlap_scaling',
    'overlap_shrink',
    'pack',
    'packmode',
    'pad',
    'page',
    'pagedir',
    'pencolor',
    'penwidth',
    'peripheries',
    'pin',
    'pos',
    'quadtree',
    'quantum',
    # 'rank',  # We intepret this one ourselves
    'rankdir',
    'ranksep',
    'ratio',
    'rects',
    'regular',
    'remincross',
    'repulsiveforce',
    'resolution',
    'root',
    'rotate',
    'rotation',
    'samehead',
    'sametail',
    'samplepoints',
    'scale',
    'searchsize',
    'sep',
    'shape',
    'shapefile',
    'showboxes',
    'sides',
    'size',
    'skew',
    'smoothing',
    'sortv',
    'splines',
    'start',
    'style',
    'stylesheet',
    'tail_lp',
    'tailclip',
    'tailhref',
    'taillabel',
    'tailport',
    'tailtarget',
    'tailtooltip',
    'tailURL',
    'target',
    'tooltip',
    'truecolor',
    'URL',
    'vertices',
    'viewport',
    'voro_margin',
    'weight',
    'width',
    'xdotversion',
    'xlabel',
    'xlp',
    'z',
]


def main(opts):
    model = load_json_file(opts.infile)

    styles = load_json_file(opts.stylesheet)

    ctx = GraphContext(styles)

    overlay_args = {
        arg: getattr(opts, arg)
        for arg in opts.overlay.arguments()
    }

    overlay = opts.overlay(ctx, **overlay_args)
    overlay.draw(opts.name, model)

    print(overlay.source())


def load_json_file(file):
    if file:
        return json.load(file)
    return {}


class GraphContext(object):
    """
    Defines a context for graph objects.

    A wrapper for the Graphviz interface to define objects via
    a dictionary structure and turn it into a graph.

    Automatically adds graphviz attributes defined in the object
    see https://graphviz.org/doc/info/attrs.html for availability.

    Attributes added to the root dictionary is added as graph attributes.
    Furthermore special "attributes" are interpreted as such::

        "nodes": {
            "node_id": {
                # Node attributes
                "classes": ["mynode"], # Nodes have the base style "node"
                #
                The ``rank`` attribute is treated differently, as it is defined
                by an id in the node object. e.g. "rank": "myrank". At the

                "rank": "myrank"
            }
        },
        "edges": [
            # An edge pointing to and from the same node
            {
                "from": "node_id",
                "to": "node_id"
                "classes": ["myedge"],  # Edges have the base style "edge"
            }
        ],
        "subgraphs": {
            "subgraph_name": {
                # Recursively treated as an individual graph
            }
        }
        "classes": ["mygraph"],  #  All graphs have the base style "graph"


        "ranks": {
            "myrank": "same",
        }
        "styles": {
            "mynode": {

            }
        }
        }
    """
    base_styles = {
        'graph': {},
        'node': {},
        'edge': {},
    }

    def __init__(
        self, stylesheet: dict, path: str = '', prefix: str = '',
        _level: int = 0
    ):
        self.graph = None
        self.styles = self.base_styles.copy()
        self.add_stylesheet(stylesheet)
        self._ranks = {}
        self._level = _level
        self.prefix = prefix
        self.path = path

    def init_graph(self, name, graph_class, attributes):
        styles = attributes.get('styles', {})

        graph_attrs = styles.get('graph', {})
        graph_attrs.update(attributes)

        graph_attrs = self._build_attributes(
            'graph',
            graph_attrs
        )
        self.graph = graph_class(
            name,
            graph_attr=graph_attrs,
            node_attr=self._build_attributes(
                'node',
                styles.get('node'),
            ),
            edge_attr=self._build_attributes(
                'edge',
                styles.get('edge'),
            ),
        )

    def add_stylesheet(self, stylesheet: dict):
        """
        Add stylesheet to the context stylesheet.

        A stylesheet consists of a dictionary of class
        definitions which are themselves dictionaries.
        """
        new_stylesheet = self.styles.copy()
        for class_name, styles in stylesheet.items():
            new_class = new_stylesheet.get(class_name, {})
            for attr, value in styles.items():
                if attr == 'style':
                    if 'style' not in new_class:
                        new_class['style'] = value
                    else:
                        for v in value:
                            if v in new_class['style']:
                                continue
                            new_class['style'].append(v)
                else:
                    new_class[attr] = value
            new_stylesheet[class_name] = new_class
        self.styles = new_stylesheet

    def new_context(self, name, path, model):
        stylesheet = self.styles.copy()
        ctx = GraphContext(
            stylesheet,
            path=path,
            prefix=model.get('prefix', ''),
            _level=self._level + 1,
        )

        if model.get('cluster', False):
            if not name.startswith('cluster_'):
                name = f'cluster_{name}'

        ctx.init_graph(name, self.graph.__class__, model)
        return ctx

    def add_subgraph_from_context(self, ctx):
        self.graph.subgraph(
            ctx.graph,
        )
        for rank, nodes in ctx.get_ranks().items():
            self._ranks[rank] = self._ranks.get(rank, []) + nodes

    def get_ranks(self):
        return self._ranks

    def node_id(self, name):
        if not self.prefix:
            return name
        return f'{self.prefix}_{name}'

    def add_edge(
        self, node_from: str, node_to: str, attributes: dict = None,
        classes=None
    ):
        classes = classes or []

        attrs = self._build_attributes(
            'edge',
            attributes,
            classes,
        )

        self.graph.edge(
            node_from,
            node_to,
            **attrs
        )

    def add_node(self, name, attributes=None, classes=None):
        classes = classes or []

        if 'rank' in attributes:
            rank_name = attributes['rank']
            rank_nodes = self._ranks.get(rank_name, [])
            self._ranks[rank_name] = rank_nodes + [name]

        attrs = self._build_attributes(
            'node',
            attributes,
            classes,
        )

        if 'label' in attrs and isinstance(attrs['label'], dict):
            attrs['label'] = format_html_label(attrs['label'])
            attrs['shape'] = 'plain'

        self.graph.node(
            self.node_id(name),
            **attrs
        )

    def _build_attributes(
        self, element_type: str, attributes: dict, classes: list = None
    ) -> dict:
        """
        Construct a dictionary of attributes for a graphviz element.

        Filters out any non-valid attributes and applies any attributes
        defined in any of the classes.

        Classes work like CSS classes, any attribute defined gets
        overriden by that later defined one.

        Any attributes specified overrides the one supplied by
        any classes.

        :param str type: The DOT element, graph, subgraph, node, edge
        :param dict attributes: (key, value) pairs. The value is
            treated as a literal.
        :param list classes: Any classes to apply defined by the stylesheet.
        :returns: A (key, value) mapping of element attributes.
        :rtype: dict
        """
        attrs = self.styles.get(element_type, {}).copy()

        classes = classes or []
        attributes = attributes or {}

        for c in classes + attributes.get('classes', []):
            attrs.update(self.styles.get(c, {}))

        if attributes:
            attrs.update(attributes)

        styles = attrs.get('style', [])

        if not isinstance(styles, list):
            raise ValueError(
                "style attribute must be a list, got '%r'" % styles
            )

        if not attributes.get('visible', True):
            styles.append('invis')
        elif attributes.get('cluster', False):
            if not styles:
                styles.append('solid')

        if styles:
            attrs['style'] = ','.join(styles)

        result = {}
        for attr in attrs.keys():
            if attr in valid_attrs:
                result[attr] = attrs[attr]
        return result

    def add_rank(self, rank_name, rank_type):
        rank_nodes = self._ranks.get(rank_name)

        fmt = '{{rank={rank_type}; {nodenames}}}'
        self.graph.body.append(
            fmt.format(
                rank_type=rank_type,
                nodenames=' '.join(rank_nodes)),
        )

    def source(self):
        return self.graph.source


def format_html_label(label):
    """Produces a graphviz html label from a dictionary."""

    rows = []
    for row in label['trs']:
        tds = []
        for td in row:
            tds.append(
                format_html_tag('td', td, td['value'])
            )
        rows.append(
            format_html_tag('tr', inner=''.join(tds))
        )
    table = format_html_tag('table', label, ''.join(rows))
    return f'<{table}>'


def format_html_tag(tag, attrs={}, inner=''):
    html_attrs = {
        'TABLE': [
            'ALIGN', 'BGCOLOR', 'BORDER', 'CELLBORDER', 'CELLPADDING',
            'CELLSPACING', 'COLOR', 'COLUMNS', 'FIXEDSIZE', 'GRADIENTANGLE',
            'HEIGHT', 'HREF', 'ID', 'PORT', 'ROWS', 'SIDES',
            'STYLE', 'TARGET', 'TITLE', 'TOOLTIP', 'VALIGN',
            'WIDTH',
        ],
        'TD': [
            'ALIGN', 'BALIGN', 'BGCOLOR', 'BORDER', 'CELLPADDING',
            'CELLSPACING', 'COLOR', 'COLSPAN', 'FIXEDSIZE', 'GRADIENTANGLE',
            'HEIGHT', 'HREF', 'ID', 'PORT', 'ROWSPAN', 'SIDES',
            'STYLE', 'TARGET', 'TITLE', 'TOOLTIP', 'VALIGN',
            'WIDTH',
        ]
    }
    tag = tag.upper()
    tag_attrs = [
        f'{k.upper()}="{v}"'
        for k, v in attrs.items()
        if k.upper() in html_attrs[tag]
    ]
    return f"<{tag} {' '.join(tag_attrs)}>{inner}</{tag}>"


overlays = [
    GraphOverlay,
    DigraphOverlay,
    EntityRelationshipOverlay,
]


def run():
    parser = ArgumentParser()
    parser.add_argument(
        '-n', '--name',
        default='G',
        help='Name of the graph')
    parser.add_argument(
        '-i', '--infile',
        type=FileType(mode='r'),
        help='Input json file',
        default=sys.stdin
    )
    parser.add_argument(
        '-s', '--stylesheet',
        type=FileType(mode='r'),
        default=None,
        help='Stylesheet file',
    )

    subparsers = parser.add_subparsers(
        title='Overlays',
        description='The overlay that will be used to generate the graph.',
        required=True,
        dest='overlay',
        help='Select the overlay to render the graph.',
    )

    for overlay in overlays:
        subparser = subparsers.add_parser(overlay.name)
        args = overlay.arguments()
        for arg, params in args.items():
            arg = arg.replace('_', '-')
            subparser.add_argument(f'--{arg}', **params)
        subparser.set_defaults(overlay=overlay)

    main(parser.parse_args())



import graphviz


class GraphOverlay(object):
    """
    Overlay generating an undirected graph.
    """
    name = 'graph'

    styles = {
        'highlighted': {
            'penwidth': '3',
        },
        'shaded': {
            'color': 'darkgrey',
            'fontcolor': 'darkgrey',
            'fillcolor': 'lightgrey',
        }
    }

    def __init__(self, ctx, select, highlight, shade, remove_deselected):
        self.ctx = ctx
        self.ctx.add_stylesheet(self.styles)
        self.selected_paths = [
            p.strip()
            for p in select.split(',')
        ]
        self.highlighted_paths = [
            p.strip()
            for p in highlight.split(',')
        ]
        self.shaded_paths = [
            p.strip()
            for p in shade.split(',')
        ]
        self.remove_deselected = remove_deselected

    @classmethod
    def arguments(self):
        return {
            'select': {'default': ''},
            'highlight': {'default': ''},
            'shade': {'default': ''},
            'remove_deselected': {
                'action': 'store_true',
                'default': False,
                'help': (
                    'Remove deselected elements as opposed to hiding them. '
                    'This will also remove not-selected subgraphs instead of '
                    'just the elements.'
                ),
            },
        }

    def draw(self, name, model, graph_class=graphviz.Graph):
        self.ctx.init_graph(
            name,
            graph_class,
            model,
        )
        processed_model = self.preprocess_model(model)
        self.walk_model(self.ctx, processed_model)

    def source(self):
        return self.ctx.source()

    def preprocess_model(self, model, current_path=''):
        """
        Preprocess the model selecting only elements in selected paths
        :returns: A filtered version of the model
        """
        new_model = model.copy()
        paths = []
        if current_path:
            paths.append(current_path)

        new_model['nodes'] = self.preprocess_nodes(
            model.get('nodes', {}),
            paths,
        )

        new_model['edges'] = self.preprocess_edges(
            model.get('edges', {}),
            paths,
        )

        new_model['subgraphs'] = self.preprocess_subgraphs(
            model.get('subgraphs', {}),
            paths,
            current_path,
        )

        return new_model

    def preprocess_nodes(self, nodes, paths):
        selected_nodes = {}
        for nodeid, node in nodes.items():
            node_paths = paths + node.get('paths', [])
            if (
                not self.in_a_selected_path(node_paths)
                and self.remove_deselected
            ):
                continue
            self.preprocess_element(node, node_paths)

            selected_nodes[nodeid] = node
        return selected_nodes

    def preprocess_edges(self, edges, paths):
        selected_edges = []
        for edge in edges:
            edge_paths = paths + edge.get('paths', [])
            if (
                not self.in_a_selected_path(edge_paths)
                and self.remove_deselected
            ):
                continue

            self.preprocess_element(edge, edge_paths)

            selected_edges.append(edge)
        return selected_edges

    def preprocess_subgraphs(self, subgraphs, paths, current_path=''):
        selected_subgraphs = {}
        for subgraph_name, subgraph in subgraphs.items():

            if subgraph_name.startswith('cluster_'):
                subgraph_name = subgraph_name[8:]
                subgraph['cluster'] = True

            subgraph_path = self.subgraph_path(subgraph_name, current_path)

            paths = [subgraph_path]

            processed_subgraph = self.preprocess_model(
                subgraph,
                current_path=subgraph_path,
            )
            self.preprocess_element(processed_subgraph, paths)

            if (
                not self.in_a_selected_path(paths)
            ):
                if self.remove_deselected:
                    if not (
                        processed_subgraph.get('nodes', False)
                        or processed_subgraph.get('edges', False)
                    ):
                        selected_subgraphs.update(
                            processed_subgraph.get('subgraphs', {})
                        )
                        return selected_subgraphs
            else:
                if not (
                    processed_subgraph.get('nodes', False)
                    or processed_subgraph.get('edges', False)
                ):
                    processed_subgraph['visible'] = False
            selected_subgraphs[subgraph_name] = processed_subgraph

        return selected_subgraphs

    def preprocess_element(self, elem, paths):
        """Generic preprocessing of graph elements.

        Applies styles based on select paths.
        """

        if not self.in_a_selected_path(paths):
            elem['visible'] = False
        else:
            elem['visible'] = True

        if self.is_highlighted(paths):
            elem['classes'] = elem.get('classes', []) + ['highlighted']

        if self.is_shaded(paths):
            elem['classes'] = elem.get('classes', []) + ['shaded']

        return elem

    def subgraph_path(self, subgraph_name, current_path):
        if current_path:
            return f'{current_path}.{subgraph_name}'
        return subgraph_name

    def is_highlighted(self, paths):
        return self.paths_in_paths(paths, self.highlighted_paths)

    def is_shaded(self, paths):
        return self.paths_in_paths(paths, self.shaded_paths)

    def in_a_selected_path(self, paths):
        """Element is in a selected path"""
        if self.selected_paths == ['']:
            return True
        return self.paths_in_paths(paths, self.selected_paths)

    def paths_in_paths(self, paths, selected_paths):
        if selected_paths == ['']:
            return False

        inverted_paths = [
            p[1:]
            for p in selected_paths
            if p.startswith('^')
        ]
        in_inverted_path = False

        for path in paths:
            if any(path.startswith(p) for p in selected_paths):
                return True

            in_inverted_path = (
                in_inverted_path
                or any(path.startswith(p) for p in inverted_paths)
            )

        return (inverted_paths and not in_inverted_path)

    def partially_selected_path(self, paths):
        """Is any current path prefix of a selected path.

        This is important for subgraphs as we want to know
        if it needs to be included or not.
        """
        if self.selected_paths == ['']:
            return True
        for path in paths:
            if any(p.startswith(path) for p in self.selected_paths):
                return True
        return False

    def walk_model(self, ctx, model):
        self.add_nodes(ctx, model.get('nodes', {}))
        self.add_subgraphs(ctx, model.get('subgraphs', {}))
        self.add_edges(ctx, model.get('edges', {}))
        self.add_ranks(ctx, model.get('ranks', {}))

    def add_nodes(self, ctx, nodes):
        for node_id, attributes in nodes.items():
            ctx.add_node(node_id, attributes)

    def add_subgraphs(self, ctx, subgraphs):
        for subgraph_name, model in subgraphs.items():
            subgraph_ctx = ctx.new_context(
                subgraph_name,
                subgraph_name,
                model,
            )

            self.walk_model(subgraph_ctx, model)

            ctx.add_subgraph_from_context(subgraph_ctx)

    def add_edges(self, ctx, edges):
        for edge in edges:
            ctx.add_edge(edge['from'], edge['to'], edge)

    def add_ranks(self, ctx, ranks):
        for rank_name, rank_type in ranks.items():
            ctx.add_rank(rank_name, rank_type)


class DigraphOverlay(GraphOverlay):
    """
    Overlay generating a directed graph.
    """

    name = 'digraph'

    def draw(self, name, model):
        super().draw(name, model, graphviz.Digraph)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['graphviz_overlay', 'graphviz_overlay.overlays']

package_data = \
{'': ['*']}

install_requires = \
['graphviz>=0.19.1,<0.20.0']

entry_points = \
{'console_scripts': ['graphviz-overlay = graphviz_overlay.graph:run']}

setup_kwargs = {
    'name': 'graphviz-overlay',
    'version': '0.1.1',
    'description': 'Generate templated graphviz source files from structured data',
    'long_description': "#####\nAbout\n#####\n\nGenerate templated graphviz source files from structured data\n\nWhy\n===\n\nThe Graphviz dot language offers a descriptive and flexible way\nto define graphs of various sorts. However, I think it also has\nsome issues.\n\n- Maintenance/small changes quickly becomes a hassle, especially\n  when you want to produce similar looking graphs. I.e. highlighting\n  certain elements, or selecting specific parts (it is possible\n  to render just subgraphs, but for me that is not always what\n  I want).\n\n- The layout engine is not exactly what you would call intuitive\n  and getting it to do your bidding takes time and effort to\n  discover. Also, you need to pull (and remember those) weird\n  tricks. Basically, it's just boilerplate work.\n\n- The defaults does not really produce pretty looking graphs, and\n  customizing the design is often tedious work and for me often\n  ends up involving a bunch of copy paste. For SVG output\n  stylesheets are supported, however that doesn't really work\n  with common cooperation tools (such as slack and github)\n\nSo, to address these issues I decided to put together a tool\nto address all that (and more).\n\n\nUsage\n=====\n\nCLI\n---\n\nThe ``graphviz-overlay`` executable reads from stdin and produces\ndot source::\n\n    cat examples/simple.json | graphviz-overlay\n\n\nVersion History\n===============\n\n0.1.1:\n  - Add initial documentation and project description\n\n0.1.0:\n  - ``graph``, ``digraph`` and ``er`` commands.\n  - Support for ``--select``, ``--highlight`` and ``--shade``\n    via paths.\n  - Support for ranks\n  - External stylesheet definition.\n  - Nodes, edges, and graphs can have `classes`.\n",
    'author': 'Ronni Elken Lindsgaard',
    'author_email': 'ronni.lindsgaard@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/rlindsgaard/graphviz-overlay',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from graphviz_overlay.overlays.graphviz import GraphOverlay

class EntityRelationshipOverlay(GraphOverlay):

    name = 'er'

    styles = {
        'entity': {
            'shape': 'box',
        },
        'attribute': {},
        'relationship': {
            'shape': 'diamond',
            'style': ['filled'],
            'color': 'lightgrey',
        },
        'cardinality': {},
    }

    def walk_model(self, ctx, model):
        self.add_domains(ctx, model.get('domains', {}))
        self.add_entity_relationships(ctx, model)
        self.add_ranks(ctx, model.get('ranks', {}))

    def add_domains(self, ctx, domains):
        for name, domain in domains.items():
            self.add_domain(ctx, name, domain)

    def add_domain(self, ctx, name, domain):
        graphname = f'cluster_{name}'

        new_ctx = ctx.new_context(graphname, name, domain)
        self.walk_model(new_ctx, domain)
        ctx.add_subgraph_from_context(new_ctx)

    def add_entity_relationships(self, ctx, model):
        self.add_entities(ctx, model.get('entities', {}))
        self.add_relationships(ctx, model.get('relationships', []))

    def add_entities(self, ctx, entities):
        for name, entity in entities.items():
            self.add_entity(ctx, name, entity)

    def add_entity(self, ctx, name, entity):
        ctx.add_node(
            name,
            entity,
            classes=['entity'],
        )
        self.add_entity_attributes(
            ctx,
            name,
            entity,
        )

    def add_entity_attributes(self, ctx, entity_name, entity):
        for attribute_name, attribute in entity.get('attributes', {}).items():
            self.add_entity_attribute(
                ctx,
                entity_name,
                attribute_name,
                attribute,
            )

    def add_entity_attribute(
        self, ctx, entity_name, attribute_name, attribute
    ):
        node_name = labels_to_nodename(entity_name, attribute_name)

        attrs = {
            'label': attribute_name,
        }
        attrs.update(attribute)

        ctx.add_node(
            node_name,
            attrs,
            classes=['attribute'],
        )
        ctx.add_edge(node_name, entity_name)

    def add_relationships(self, ctx, relationships):
        for relationship in relationships:
            self.add_relationship(ctx, relationship)

    def add_relationship(self, ctx, relationship):
        label = relationship.get(
            'label',
            '-'.join([
                relationship['from']['name'].upper()[0],
                relationship['to']['name'].upper()[0],
            ]))

        attrs = {
            'label': label,
        }
        attrs.update(relationship)

        ctx.add_node(
            connectorid(relationship),
            attrs,
            classes=['relationship'],
        )
        self.add_cardinality(ctx, relationship)

    def add_cardinality(self, ctx, relationship):
        connector_id = connectorid(relationship)

        from_attrs = {'label': relationship['from'].get('cardinality', '')}
        from_attrs.update(relationship['from'])
        ctx.add_edge(
            label_to_nodename(relationship['from']['name']),
            connector_id,
            from_attrs,
            classes=['cardinality'],
        )
        to_attrs = {'label': relationship['to'].get('cardinality', '')}
        to_attrs.update(relationship['to'])
        ctx.add_edge(
            connector_id,
            label_to_nodename(relationship['to']['name']),
            to_attrs,
            classes=['cardinality'],
        )


def connectorid(relationship):
    return '_'.join(
        [
            label_to_nodename(relationship['from']['name']),
            label_to_nodename(relationship['to']['name']),
        ]
    )


def labels_to_nodename(*labels):
    return '_'.join(
        label_to_nodename(label)
        for label in labels
    )


def label_to_nodename(label):
    return label.lower().replace(' ', '_').replace('-', '_')
